/******************************** LICENSE ********************************

 Copyright 2007 European Centre for Medium-Range Weather Forecasts (ECMWF)

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 ******************************** LICENSE ********************************/

/*! \file PolyShadingMethod.h
    \brief Definition of the Template class PolyShadingMethod.
    
    Magics Team - ECMWF 2004
    
    Started: Wed 18-Aug-2004
    
    Changes:
    
*/


#include "DotPolyShadingMethod.h"
#include "HatchPolyShadingMethod.h"
#include "LevelSelection.h"

using namespace magics;

struct LegendEntryBuilder
{
    LegendEntryBuilder(LegendVisitor& legend, const PolyShadingMethod& method,
    	const ColourTechnique& colours) :
    	legend_(legend), method_(method), colours_(colours)  {};
    ~LegendEntryBuilder() {};
    bool operator()(const pair<double, ColourInfo>& first, const pair<double, ColourInfo>& second) {
        Polyline* box = new Polyline();

        double min = first.second.level_;
        double max = second.second.level_;

        method_(*box, max, colours_);
        box->setFillColour(colours_.right(min));
        box->setFilled(true);
        box->setStroke(true);
        box->setColour(Colour("black"));
        legend_.add(new BoxEntry(min, max, box));
        return false;

    }
    LegendVisitor& legend_;
    const PolyShadingMethod& method_;
    const ColourTechnique& colours_;
};

	
    
void PolyShadingMethod::operator()(Polyline& poly, double, const ColourTechnique&) const
    {
       
      
        poly.setFilled(true);      
        poly.setStroke(false);      
        FillShadingProperties* shading = new FillShadingProperties();                      
        poly.setShading(shading);
    };
   
void  PolyShadingMethod::visit(LegendVisitor& legend, const ColourTechnique& colour) {
        
        MagLog::dev() << "Create legend information"  << "\n";
       
        std::adjacent_find(colour.begin(), colour.end(), LegendEntryBuilder(legend, *this, colour));
     
        legend.last(); // Flag the last entry as being the last! To get a nice labelling in countinuous mode!!!
        
    }
    
void DotPolyShadingMethod::prepare(const LevelSelection& levels) {
		if (levels.empty() )return;
      float step = (max_density_ - min_density_)/(levels.size() - 1);

          	LevelSelection::const_iterator from = levels.begin();
          	LevelSelection::const_iterator level = levels.begin();
          	level++;
          	float density = min_density_;
          	for (  ;  level != levels.end(); ++level) {
          		indexes_.insert(make_pair(Interval(*from, *level), density));

          		from++;
          		density+=step;

          	}
          }


void  DotPolyShadingMethod::operator()(Polyline& poly, const ColourTechnique& colours) const {
    DotShadingProperties* shading = new DotShadingProperties();

    double value = poly.front().value();
    Colour colour = poly.getFillColour();
    Colour left = colours.left(value);
    Colour right = colours.right(value);



    shading->size_ = size_;
    shading->angle_ = angle_;
    shading->density_ =  indexes_.find(value, 40.);

	poly.setStroke(false);
    poly.setShading(shading);
    //MagLog::dev() << "Attach DotShading Information" << *shading << "\n";
}

void  DotPolyShadingMethod::operator()(Polyline& poly, double value, const ColourTechnique& colours) const {
	static int i = 0;

	i++;

	 poly.setFilled(true);
	  Colour colour = poly.getFillColour();
	        Colour left = colours.left(value);
	        Colour right = colours.right(value);

	        DotShadingProperties* shading = new DotShadingProperties();

	        shading->size_ = size_;
	        shading->angle_ = angle_;





    	  shading->density_ =    indexes_.find(value, 40.);


    poly.setShading(shading);
	poly.setStroke(false);
    //MagLog::dev() << "Attach DotShading Information [" << i << "]=" << value << "--->" << *shading << "\n";
}

void HatchPolyShadingMethod::prepare(const LevelSelection& levels) {
   	int index = 1;
   	if (index_ >= 7 || index_ <0) {
   		MagLog::warning() << "index should be < 7--> reset to 1 "<< endl;
   		index_ = 1 ;
   	}
   	LevelSelection::const_iterator from = levels.begin();
   	LevelSelection::const_iterator level = levels.begin();
   	level++;
   	for (  ;  level != levels.end(); ++level) {
   		indexes_.insert(make_pair(Interval(*from, *level), (index_) ? index_ : index));
   		index++;
   		from++;
   		if ( index == 7 ) index = 1;
   	}
   }

void HatchPolyShadingMethod::operator()(Polyline& poly, const ColourTechnique&) const {
       float value = poly.front().value();

       HatchShadingProperties* shading = new HatchShadingProperties();


       shading->index_     = indexes_.find(value, 1);

       MagLog::dev() << "index for " << value << "--->" << shading->index_ << "\n";
       shading->density_   = density_;
       shading->thickness_ = thickness_;
       poly.setFilled(true);
       poly.setStroke(false);
       poly.setShading(shading);
        MagLog::dev() << "Attach HatchShading Information : "  << value << "--->" << *shading << "\n";
   }
void HatchPolyShadingMethod::operator()(Polyline& poly, double value, const ColourTechnique& ) const {

       HatchShadingProperties* shading = new HatchShadingProperties();
        poly.setFilled(true);
        shading->index_     = indexes_.find(value, 1);
       shading->density_   = density_;
       shading->thickness_ = thickness_;
       poly.setStroke(false);
       if (!value)  poly.setShading(new FillShadingProperties());
       poly.setShading(shading);
        //MagLog::dev() << "Attach HatchShading Information--->" << *shading << "\n";
   }
